#ifndef _BLUEDBM_DEV_RAMSSD_H
#define _BLUEDBM_DEV_RAMSSD_H

#include <linux/module.h>
#include <linux/slab.h>
#include <linux/hrtimer.h>
#include <linux/ktime.h>

#include "bdbm_drv.h"
#include "params.h"
#include "utils/time.h"


enum RAMSSD_TIMING {
	RAMSSD_TIMING_DISABLE = 1,
	RAMSSD_TIMING_ENABLE_TASKLET,
	RAMSSD_TIMING_ENABLE_HRTIMER,
};

struct dev_ramssd_punit {
	void* ptr_req;
	int64_t target_elapsed_time_us;
	struct bdbm_stopwatch sw;
};

struct dev_ramssd_info {
	uint8_t is_init; /* 0: not initialized, 1: initialized */
	uint8_t timing_mode;	/* 0: disable timing emulation, 1: enable timing emulation */
	struct nand_params* nand_params;
	void* ptr_ssdram; /* DRAM memory for SSD */
	struct dev_ramssd_punit* ptr_punits;	/* parallel units */
	spinlock_t ramssd_lock;
	void (*intr_handler) (void*);

	struct tasklet_struct* tasklet;	/* tasklet */
	struct hrtimer hrtimer;	/* hrtimer must be at the end of the structure */
};

struct dev_ramssd_info* dev_ramssd_create (struct nand_params* ptr_nand_params, uint8_t timing_emul, void (*intr_handler)(void*));
void dev_ramssd_destroy (struct dev_ramssd_info* ptr_ramssd_info);
uint32_t dev_ramssd_send_cmd (struct dev_ramssd_info* ptr_ramssd_info, struct bdbm_llm_req_t* ptr_llm_req );
void dev_ramssd_summary (struct dev_ramssd_info* ptr_ramssd_info);

/* for snapshot */
uint32_t dev_ramssd_load (struct dev_ramssd_info* ptr_ramssd_info, const char* fn);
uint32_t dev_ramssd_store (struct dev_ramssd_info* ptr_ramssd_info, const char* fn);

/* some inline functions */
inline static 
uint64_t dev_ramssd_get_page_size_main (struct dev_ramssd_info* ptr_ramssd_info) {
	return ptr_ramssd_info->nand_params->page_main_size;
}

inline static 
uint64_t dev_ramssd_get_page_size_oob (struct dev_ramssd_info* ptr_ramssd_info) {
	return ptr_ramssd_info->nand_params->page_oob_size;
}

inline static 
uint64_t dev_ramssd_get_page_size (struct dev_ramssd_info* ptr_ramssd_info) {
	return ptr_ramssd_info->nand_params->page_main_size +
		ptr_ramssd_info->nand_params->page_oob_size;
}

inline static 
uint64_t dev_ramssd_get_block_size (struct dev_ramssd_info* ptr_ramssd_info) {
	return dev_ramssd_get_page_size (ptr_ramssd_info) *	
		ptr_ramssd_info->nand_params->nr_pages_per_block;
}

inline static 
uint64_t dev_ramssd_get_chip_size (struct dev_ramssd_info* ptr_ramssd_info) {
	return dev_ramssd_get_block_size (ptr_ramssd_info) * 
		ptr_ramssd_info->nand_params->nr_blocks_per_chip;
}

inline static 
uint64_t dev_ramssd_get_channel_size (struct dev_ramssd_info* ptr_ramssd_info) {
	return dev_ramssd_get_chip_size (ptr_ramssd_info) * 
		ptr_ramssd_info->nand_params->nr_chips_per_channel;
}

inline static 
uint64_t dev_ramssd_get_ssd_size (struct dev_ramssd_info* ptr_ramssd_info) {
	return dev_ramssd_get_channel_size (ptr_ramssd_info) * 
		ptr_ramssd_info->nand_params->nr_channels;
}

inline static 
uint64_t dev_ramssd_get_pages_per_block (struct dev_ramssd_info* ptr_ramssd_info) {
	return ptr_ramssd_info->nand_params->nr_pages_per_block;
}

inline static 
uint64_t dev_ramssd_get_blocks_per_chips (struct dev_ramssd_info* ptr_ramssd_info) {
	return ptr_ramssd_info->nand_params->nr_blocks_per_chip;
}

inline static 
uint64_t dev_ramssd_get_chips_per_channel (struct dev_ramssd_info* ptr_ramssd_info) {
	return ptr_ramssd_info->nand_params->nr_chips_per_channel;
}

inline static 
uint64_t dev_ramssd_get_channles_per_ssd (struct dev_ramssd_info* ptr_ramssd_info) {
	return ptr_ramssd_info->nand_params->nr_channels;
}

inline static 
uint64_t dev_ramssd_get_chips_per_ssd (struct dev_ramssd_info* ptr_ramssd_info) {
	return ptr_ramssd_info->nand_params->nr_channels *
		ptr_ramssd_info->nand_params->nr_chips_per_channel;
}

inline static 
uint64_t dev_ramssd_get_blocks_per_ssd (struct dev_ramssd_info* ptr_ramssd_info) {
	return dev_ramssd_get_chips_per_ssd (ptr_ramssd_info) *
		ptr_ramssd_info->nand_params->nr_blocks_per_chip;
}

inline static 
uint64_t dev_ramssd_get_pages_per_ssd (struct dev_ramssd_info* ptr_ramssd_info) {
	return dev_ramssd_get_blocks_per_ssd (ptr_ramssd_info) *
		ptr_ramssd_info->nand_params->nr_pages_per_block;
}

inline static 
uint8_t dev_ramssd_is_init (struct dev_ramssd_info* ptr_ramssd_info) {
	if (ptr_ramssd_info->is_init == 1)
		return 0;
	else 
		return 1;
}

#endif /* _BLUEDBM_DEV_RAMSSD_H */
